package mvd.jester.model;

import java.util.Set;

/**
 * Task
 */
public class SynchronousTask implements Task {

    private final long deadline;
    private final long period;
    private final Set<Segment> segments;
    private final long workload;
    private final long criticalPath;
    private final long numberOfThreads;

    public SynchronousTask(final long period, final long deadline, final long numberOfThreads,
            final Set<Segment> segments) {
        this.deadline = deadline;
        this.period = period;
        this.numberOfThreads = numberOfThreads;
        this.segments = segments;
        this.workload = SynchronousUtils.calculateWorkload(segments);
        this.criticalPath = SynchronousUtils.calculateCriticalPath(segments);
    }

    public SynchronousTask(final Set<Segment> segments, final long period,
            final long numberOfThreads) {
        this(period, period, numberOfThreads, segments);
    }

    public double getUtilization() {
        return (double) workload / period;
    }

    /**
     * @return the deadline
     */
    public long getDeadline() {
        return deadline;
    }

    /**
     * @return the period
     */
    public long getPeriod() {
        return period;
    }

    /**
     * @return the segments
     */
    public Set<Segment> getWorkloadDistribution() {
        return segments;
    }

    /**
     * @return the maximumWcet
     */
    public long getWorkload() {
        return workload;
    }

    /**
     * @return the criticalPath
     */
    public long getCriticalPath() {
        return criticalPath;
    }

    @Override
    public long getMaximumParallelism() {
        long max = 0;
        for (final Segment s : segments) {
            if (max < s.getNumberOfJobs()) {
                max = s.getNumberOfJobs();
            }
        }
        return max;
    }

    @Override
    public long getNumberOfThreads() {
        return numberOfThreads;
    }

    public static class SynchronousUtils {
        public static long calculateWorkload(final Set<Segment> segments) {
            long workload = 0;
            for (final Segment s : segments) {
                workload += s.getJobWcet() * s.getNumberOfJobs();
            }
            return workload;
        }


        public static long calculateCriticalPath(final Set<Segment> segments) {
            long criticalPath = 0;
            for (final Segment s : segments) {
                criticalPath += s.getJobWcet();
            }

            return criticalPath;
        }
    }
}
