package mvd.jester.simulator;

import java.util.HashSet;
import java.util.List;
import java.util.Optional;
import java.util.Set;
import java.util.TreeSet;
import java.util.stream.Collectors;
import org.jgrapht.Graphs;
import mvd.jester.model.DagTask;
import mvd.jester.model.Subtask;
import mvd.jester.priority.PriorityManager;
import mvd.jester.simulator.events.*;
import mvd.jester.simulator.exceptions.DeadlineMissedException;
import mvd.jester.simulator.exceptions.SchedulingException;
import mvd.jester.simulator.model.ExtendedSubtaskContext;
import mvd.jester.simulator.model.JobContext;
import mvd.jester.simulator.model.ProcessorContext;
import mvd.jester.simulator.model.ProcessorContext.ProcessorComparator;

public class GlobalScheduler {

    private final int nbProcessors;
    private final PriorityManager priorityManager;

    public GlobalScheduler(final PriorityManager priorityManager, final int nbProcessors) {
        this.priorityManager = priorityManager;
        this.nbProcessors = nbProcessors;
    }

    public void schedule(final Set<DagTask> taskSet, final long duration) {
        final SchedulingContext sc = new SchedulingContext(priorityManager, nbProcessors, duration);
        taskSet.forEach(t -> sc.events.offer(new TaskReleasedEvent(0, t)));
        while (!sc.events.isEmpty()) {
            final IEvent event = sc.events.poll();
            updateExecution(sc, event.getTime());
            handle(event, sc);
        }
    }

    private void handle(final IEvent event, final SchedulingContext sc) {
        event.log();
        switch (event.getType()) {
            case SUBTASK_ACTIVATED:
                handleSubtaskActivated((SubtaskActivatedEvent) event, sc);
                break;
            case SUBTASK_COMPLETED:
                handleSubtaskCompleted((SubtaskCompletedEvent) event, sc);
                break;
            case SUBTASK_CHECK_COMPLETED:
                handleSubtaskCheckCompleted((SubtaskCheckCompletedEvent) event, sc);
                break;
            case SUBTASK_PREEMPTED:
                handleSubtaskPreempted((SubtaskPreemptedEvent) event, sc);
                break;
            case SUBTASK_STARTED:
                handleSubtaskStarted((SubtaskStartedEvent) event, sc);
                break;
            case SUBTASK_STOPPED:
                handleSubtaskStopped((SubtaskStoppedEvent) event, sc);
                break;
            case JOB_ACTIVATED:
                handleJobActivated((JobActivatedEvent) event, sc);
                break;
            case JOB_COMPLETED:
                handleJobCompleted((JobCompletedEvent) event, sc);
                break;
            case TASK_RELEASED:
                handleTaskReleased((TaskReleasedEvent) event, sc);
                break;
            default:
                throw new IllegalArgumentException(
                        event.getType() + " is not handled by " + this.getClass());
        }
    }

    private void handleJobCompleted(final JobCompletedEvent event, final SchedulingContext sc) {
        if (event.getTime() > event.getJobContext().getDeadline()) {
            throw new DeadlineMissedException(event.getJobContext() + " misses its deadline");
        }
    }

    private void handleSubtaskActivated(final SubtaskActivatedEvent event,
            final SchedulingContext sc) {
        final Subtask subtask = event.getSubtask();
        final JobContext job = event.getJobContext();
        sc.waitingQueue.offer(new ExtendedSubtaskContext(job, subtask));
        updateAdmission(sc, event.getTime());
    }

    private void handleSubtaskCheckCompleted(final SubtaskCheckCompletedEvent event,
            final SchedulingContext sc) {
        final ProcessorContext processor = event.getProcessor();
        if (processor.isIdle()) {
            throw new SchedulingException(
                    processor + " should execute " + event.getSubtaskContext());
        }
        final Optional<ExtendedSubtaskContext> context = processor.getSubtask();
        if (context.isPresent() && context.get().getRemainingExecutionTime() == 0) {
            sc.events.offer(new SubtaskStoppedEvent(event.getTime(), context.get(), processor));
        }
    }

    private void handleSubtaskCompleted(final SubtaskCompletedEvent event,
            final SchedulingContext sc) {
        final ExtendedSubtaskContext subtask = event.getSubtaskContext();
        final JobContext job = subtask.getJobContext();

        job.registerFinishedSubtask(subtask);
        final DagTask task = subtask.getJobContext().getTask();
        final List<Subtask> successor =
                Graphs.successorListOf(task.getJobDag(), subtask.getSubtask());
        if (successor.isEmpty()) {
            sc.events.offer(new JobCompletedEvent(event.getTime(), job));
            updateAdmission(sc, event.getTime());
        } else {
            final List<Subtask> finishedSubtasks = job.getFinishedSubtasks().stream()
                    .map(esc -> esc.getSubtask()).collect(Collectors.toList());
            for (final Subtask s : successor) {
                final List<Subtask> predecessors = Graphs.predecessorListOf(task.getJobDag(), s);
                if (finishedSubtasks.containsAll(predecessors)) {
                    sc.events.offer(new SubtaskActivatedEvent(event.getTime(), job, s));
                }
            }
        }
    }

    private void handleSubtaskPreempted(final SubtaskPreemptedEvent event,
            final SchedulingContext sc) {
        final ProcessorContext processor = event.getProcessor();
        final long time = event.getTime();
        sc.events.offer(new SubtaskStartedEvent(time, event.getNextSubtask(), processor));
        sc.events.offer(new SubtaskStoppedEvent(time, event.getSubtaskContext(), processor));
    }

    private void handleSubtaskStarted(final SubtaskStartedEvent event, final SchedulingContext sc) {
        final long time = event.getTime();
        final ProcessorContext processor = event.getProcessor();
        final ExtendedSubtaskContext context = event.getSubtaskContext();
        if (processor.accept(context, time)) {
            sc.events.offer(new SubtaskCheckCompletedEvent(
                    time + context.getRemainingExecutionTime(), context, processor));
        }
    }

    private void handleSubtaskStopped(final SubtaskStoppedEvent event, final SchedulingContext sc) {
        final ExtendedSubtaskContext context = event.getSubtaskContext();
        final ProcessorContext processor = event.getProcessor();
        if (context.getRemainingExecutionTime() > 0) {
            sc.waitingQueue.offer(context);
        } else {
            sc.events.offer(new SubtaskCompletedEvent(event.getTime(), context));
        }
        processor.free();
    }

    private void handleJobActivated(final JobActivatedEvent event, final SchedulingContext sc) {
        final JobContext job = event.getJobContext();

        final DagTask task = job.getTask();
        final long nextRelease = event.getTime() + task.getPeriod();
        if (nextRelease < sc.duration) {
            final JobContext nextJob = new JobContext(task, nextRelease);
            sc.events.offer(new JobActivatedEvent(nextRelease, nextJob));
        }

        sc.events.offer(new SubtaskActivatedEvent(event.getTime(), job, job.getSource()));
    }

    private void handleTaskReleased(final TaskReleasedEvent event, final SchedulingContext sc) {
        final DagTask task = event.getTask();
        final long release = event.getTime();
        final JobContext job = new JobContext(task, release);
        sc.events.offer(new JobActivatedEvent(release, job));
    }

    private void updateAdmission(final SchedulingContext sc, final long time) {
        Set<ProcessorContext> sortedProcessors =
                new TreeSet<>(new ProcessorComparator(priorityManager));
        sortedProcessors.addAll(sc.processors);
        for (final ProcessorContext processor : sortedProcessors) {
            if (processor.canAccept(sc.waitingQueue.peek(), priorityManager)) {
                final ExtendedSubtaskContext context = sc.waitingQueue.poll();
                if (processor.isIdle()) {
                    sc.events.offer(new SubtaskStartedEvent(time, context, processor));
                } else {
                    sc.events.offer(new SubtaskPreemptedEvent(time, processor.getSubtask().get(),
                            processor, context));
                }
            }
        }
    }

    private void updateExecution(final SchedulingContext sc, final long time) {
        for (final ProcessorContext processor : sc.processors) {
            processor.update(time);
        }
    }

    private class SchedulingContext {

        private final EventQueue events;
        private final Set<ProcessorContext> processors;
        private final WaitQueue waitingQueue;
        private final long duration;

        private SchedulingContext(final PriorityManager priorityManager, final int nbProcessors,
                final long duration) {
            this.events = new EventQueue(priorityManager);
            this.waitingQueue = new WaitQueue(priorityManager);
            this.processors = new HashSet<>();
            for (int m = 0; m < nbProcessors; ++m) {
                processors.add(new ProcessorContext(m));
            }
            this.duration = duration;
        }

    }
}
