package mvd.jester.utils;

import java.util.Arrays;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedHashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Optional;
import java.util.Set;
import com.google.common.collect.Iterables;
import com.google.common.collect.Sets;
import org.jgrapht.Graphs;
import org.jgrapht.experimental.dag.DirectedAcyclicGraph;
import org.jgrapht.graph.DefaultEdge;
import org.jgrapht.traverse.TopologicalOrderIterator;
import mvd.jester.model.Subtask;
import mvd.jester.model.Segment;
import mvd.jester.utils.BinaryDecompositionTree.Node;
import mvd.jester.utils.BinaryDecompositionTree.NodeType;

public class DagUtils {

    public static long calculateWorkload(final DirectedAcyclicGraph<Subtask, DefaultEdge> jobDag) {
        long workload = 0;

        for (final Subtask job : jobDag) {
            workload += job.getWcet();
        }
        return workload;
    }

    public static long calculateCriticalPath(final DirectedAcyclicGraph<Subtask, DefaultEdge> jobDag) {
        long criticalPath = 0;
        for (final Subtask job : jobDag) {
            final Set<DefaultEdge> edges = jobDag.incomingEdgesOf(job);
            long longestRelativeCompletionTime = 0;
            for (final DefaultEdge e : edges) {
                final Subtask source = jobDag.getEdgeSource(e);
                longestRelativeCompletionTime =
                        longestRelativeCompletionTime >= source.getRelativeCompletionTime()
                                ? longestRelativeCompletionTime
                                : source.getRelativeCompletionTime();
            }

            job.setRelativeCompletionTime(longestRelativeCompletionTime + job.getWcet());
            criticalPath = job.getRelativeCompletionTime();
        }

        return criticalPath;
    }

    public static LinkedHashSet<Segment> calculateWorkloadDistribution(
            final DirectedAcyclicGraph<Subtask, DefaultEdge> jobDag, final long criticalPath) {
        final LinkedHashSet<Segment> segments = new LinkedHashSet<>();
        long segmentDuration = 0;
        long segmentHeight = 1;
        for (long t = 0; t < criticalPath; ++t) {
            long currentHeight = 0;
            for (final Subtask j : jobDag) {
                if (t >= j.getRelativeCompletionTime() - j.getWcet()
                        && t < j.getRelativeCompletionTime()) {
                    currentHeight++;
                }
            }
            if (currentHeight == segmentHeight) {
                segmentDuration++;
            } else {
                segments.add(new Segment(segmentDuration, segmentHeight));
                segmentDuration = 1;
                segmentHeight = currentHeight;
            }
        }
        segments.add(new Segment(segmentDuration, segmentHeight));

        return segments;
    }

    // public static DirectedAcyclicGraph<Job, DefaultEdge> createNFJGraph(
    // final DirectedAcyclicGraph<Job, DefaultEdge> jobDag) {
    // final LinkedList<Job> joinNodes = new LinkedList<>();
    // final LinkedList<Job> forkNodes = new LinkedList<>();

    // collectForksAndJoins(jobDag, forkNodes, joinNodes);
    // return createNFJGraph(jobDag, forkNodes, joinNodes);
    // }

    public static void collectForksAndJoins(final DirectedAcyclicGraph<Subtask, DefaultEdge> jobDag,
            final LinkedList<Subtask> forkNodes, final LinkedList<Subtask> joinNodes) {
        // final ClosestFirstIterator<Job, DefaultEdge> it = new ClosestFirstIterator<>(jobDag);
        Iterator<Subtask> it = jobDag.iterator();
        for (; it.hasNext();) {
            final Subtask j = it.next();
            // for (final Job j : jobDag) {
            if (jobDag.inDegreeOf(j) > 1) {
                joinNodes.add(j);
            }
            if (jobDag.outDegreeOf(j) > 1) {
                forkNodes.add(j);
            }
        }
    }



    private static LinkedList<Subtask> getUnvisitedJoinNodes(
            final DirectedAcyclicGraph<Subtask, DefaultEdge> jobDag, final Set<Subtask> visitedJobs) {
        final LinkedList<Subtask> joinNodes = new LinkedList<>();
        final TopologicalOrderIterator<Subtask, DefaultEdge> it =
                new TopologicalOrderIterator<>(jobDag);
        for (; it.hasNext();) {
            final Subtask j = it.next();
            if (jobDag.inDegreeOf(j) > 1 && !visitedJobs.contains(j)) {
                joinNodes.add(j);
            }
        }
        return joinNodes;
    }

    public static DirectedAcyclicGraph<Subtask, DefaultEdge> createNFJGraph(
            final DirectedAcyclicGraph<Subtask, DefaultEdge> jobDag) {
        final DirectedAcyclicGraph<Subtask, DefaultEdge> modifiedJobDag =
                new DirectedAcyclicGraph<>(DefaultEdge.class);
        Graphs.addGraph(modifiedJobDag, jobDag);
        Subtask sink = null;

        for (final Subtask j : modifiedJobDag) {
            sink = j;
        }

        final Set<Subtask> visitedJobs = new HashSet<>();
        while (true) {
            final LinkedList<Subtask> joins = getUnvisitedJoinNodes(modifiedJobDag, visitedJobs);
            if (joins.isEmpty()) {
                break;
            }
            final Subtask joinNode = joins.getFirst();
            visitedJobs.add(joinNode);

            nextEdge: while (modifiedJobDag.inDegreeOf(joinNode) > 1) {
                final List<Subtask> predecessors = Graphs.predecessorListOf(modifiedJobDag, joinNode);
                for (final Subtask p : predecessors) {
                    if (!checkPredecessors(modifiedJobDag, joinNode, p)) {
                        final DefaultEdge conflictingEdge = modifiedJobDag.getEdge(p, joinNode);
                        modifiedJobDag.removeEdge(conflictingEdge);
                        if (modifiedJobDag.outDegreeOf(p) == 0) {
                            try {
                                modifiedJobDag.addDagEdge(p, sink);
                            } catch (final Exception exception) {

                            }
                        }
                        continue nextEdge;
                    }
                }
                break;
            }
        }

        return modifiedJobDag;

    }

    private static boolean checkPredecessors(final DirectedAcyclicGraph<Subtask, DefaultEdge> jobDag,
            final Subtask joinNode, Subtask current) {
        if (jobDag.outDegreeOf(current) > 1) {
            final Set<Subtask> ancestorsOfJ = jobDag.getAncestors(jobDag, joinNode);
            ancestorsOfJ.add(joinNode);
            if (!ancestorsOfJ.containsAll(Graphs.successorListOf(jobDag, current))) {
                return false;
            }

            final Set<Subtask> descendantsOfCurrent = jobDag.getDescendants(jobDag, current);
            if (descendantsOfCurrent.containsAll(Graphs.predecessorListOf(jobDag, joinNode))) {
                return true;
            }
        }
        final List<Subtask> predecessorList = Graphs.predecessorListOf(jobDag, current);
        for (Subtask p : predecessorList) {
            if (!checkPredecessors(jobDag, joinNode, p)) {
                return false;
            }
        }
        return true;
    }

    public static BinaryDecompositionTree<Subtask> createDecompositionTree(
            final DirectedAcyclicGraph<Subtask, DefaultEdge> jobDag) {
        final LinkedList<Subtask> joinNodes = new LinkedList<>();
        final LinkedList<Subtask> forkNodes = new LinkedList<>();

        collectForksAndJoins(jobDag, forkNodes, joinNodes);
        return createDecompositionTree(jobDag, forkNodes, joinNodes);
    }

    public static BinaryDecompositionTree<Subtask> createDecompositionTree(
            final DirectedAcyclicGraph<Subtask, DefaultEdge> jobDag, final LinkedList<Subtask> forkNodes,
            final LinkedList<Subtask> joinNodes) {
        final BinaryDecompositionTree<Subtask> tree = new BinaryDecompositionTree<>();
        Optional<Subtask> source = Optional.empty();
        Optional<Subtask> sink = Optional.empty();
        boolean firstRun = true;
        for (final Subtask j : jobDag) {
            if (firstRun) {
                firstRun = false;
                source = Optional.of(j);
            }
            sink = Optional.of(j);
        }

        if (source.isPresent() && sink.isPresent()) {
            final Subtask current = source.get();

            DagUtils.constructTree(jobDag, null, current, tree, forkNodes, joinNodes, sink.get());
        }

        return tree;
    }

    private static Node<Subtask> constructTree(final DirectedAcyclicGraph<Subtask, DefaultEdge> jobDag,
            final Node<Subtask> parentNode, final Subtask currentJob,
            final BinaryDecompositionTree<Subtask> tree, final LinkedList<Subtask> forkVertices,
            final LinkedList<Subtask> joinVertices, final Subtask sinkJob) {
        if (forkVertices.contains(currentJob)) {
            final Subtask forkJob = currentJob;
            final Subtask joinJob = findJoin(jobDag, forkJob, sinkJob, joinVertices);

            final Node<Subtask> seqForkNode = new Node<Subtask>(parentNode, NodeType.SEQ);
            seqForkNode.setLeftNode(new Node<Subtask>(seqForkNode, forkJob));
            if (tree.isEmpty()) {
                tree.setRoot(seqForkNode);
            }
            final Node<Subtask> parContinuationNode;
            if (!tree.contains(joinJob)) {
                final Node<Subtask> seqJoinNode =
                        seqForkNode.setRightNode(new Node<Subtask>(seqForkNode, NodeType.SEQ));

                final Node<Subtask> subTreeOfJoin = constructTree(jobDag, seqJoinNode, joinJob, tree,
                        forkVertices, joinVertices, sinkJob);
                seqJoinNode.setRightNode(subTreeOfJoin);

                parContinuationNode =
                        seqJoinNode.setLeftNode(new Node<Subtask>(seqJoinNode, NodeType.PAR));
            } else {
                parContinuationNode =
                        seqForkNode.setRightNode(new Node<Subtask>(seqForkNode, NodeType.PAR));
            }


            Node<Subtask> successorParent = parContinuationNode;
            final List<Subtask> successors = Graphs.successorListOf(jobDag, currentJob);
            // create leftSide of joinNode
            for (int i = 0; i < successors.size(); ++i) {
                final Subtask succ = successors.get(i);
                final Node<Subtask> thisNode = constructTree(jobDag, successorParent, succ, tree,
                        forkVertices, joinVertices, sinkJob);
                if (i == successors.size() - 1) {
                    successorParent.setRightNode(thisNode);
                } else if (i == successors.size() - 2) {
                    successorParent.setLeftNode(thisNode);
                } else {
                    successorParent.setLeftNode(thisNode);
                    successorParent.setRightNode(new Node<>(successorParent, NodeType.PAR));
                    successorParent = successorParent.getRightNode();
                }
            }

            return seqForkNode;
        } else {
            final List<Subtask> successorList = Graphs.successorListOf(jobDag, currentJob);
            if (successorList.size() > 0) {
                final Subtask successor = successorList.get(0);
                if (!tree.contains(successor)) {
                    final Node<Subtask> seqJobNode = new Node<Subtask>(parentNode, NodeType.SEQ);
                    if (tree.isEmpty()) {
                        tree.setRoot(seqJobNode);
                    }
                    seqJobNode.setLeftNode(new Node<Subtask>(seqJobNode, currentJob));
                    final Node<Subtask> contNode = constructTree(jobDag, seqJobNode, successor, tree,
                            forkVertices, joinVertices, sinkJob);
                    seqJobNode.setRightNode(contNode);
                    return seqJobNode;
                }
            }
            final Node<Subtask> jobNode = new Node<Subtask>(parentNode, currentJob);
            if (tree.isEmpty()) {
                tree.setRoot(jobNode);
            }
            return jobNode;
        }
    }



    private static Subtask findJoin(final DirectedAcyclicGraph<Subtask, DefaultEdge> jobDag,
            final Subtask forkNode, final Subtask sink, final List<Subtask> joinNodes) {
        for (final Subtask j : joinNodes) {
            final Set<Subtask> ancestorsOfJ = jobDag.getAncestors(jobDag, j);
            final Set<Subtask> ancestorsOfFork = jobDag.getAncestors(jobDag, forkNode);
            ancestorsOfFork.add(forkNode);
            final Set<Subtask> inbetweenJobs = new HashSet<>(ancestorsOfJ);
            inbetweenJobs.removeAll(ancestorsOfFork);
            inbetweenJobs.add(j);
            if (inbetweenJobs.isEmpty()) {
                continue;
            }

            final List<Subtask> successorOfFork = Graphs.successorListOf(jobDag, forkNode);

            if (inbetweenJobs.containsAll(successorOfFork)) {
                return j;
            }
        }

        return sink;
    }


    public static boolean checkNFJProperty(final DirectedAcyclicGraph<Subtask, DefaultEdge> jobDag) {
        final LinkedList<Subtask> joinNodes = new LinkedList<>();
        final LinkedList<Subtask> forkNodes = new LinkedList<>();

        collectForksAndJoins(jobDag, forkNodes, joinNodes);

        if (joinNodes.isEmpty()) {
            return true;
        }

        nextJoin: for (final Subtask j : joinNodes) {
            nextFork: for (final Subtask f : forkNodes) {
                final Set<Subtask> ancestorsOfJ = jobDag.getAncestors(jobDag, j);
                final Set<Subtask> descendantsOfF = jobDag.getDescendants(jobDag, f);
                final Set<Subtask> inbetweenJobs = new HashSet<>(ancestorsOfJ);
                inbetweenJobs.retainAll(descendantsOfF);
                if (inbetweenJobs.isEmpty()) {
                    continue;
                }

                for (final Subtask a : inbetweenJobs) {
                    final List<Subtask> neighboursOfA = Graphs.neighborListOf(jobDag, a);
                    for (final Subtask b : neighboursOfA) {
                        if ((jobDag.getAncestors(jobDag, j).contains(b) || b == j)
                                && (jobDag.getDescendants(jobDag, f).contains(b) || b == f)) {
                            // this b is ok
                        } else {
                            continue nextFork;
                        }
                    }
                }
                continue nextJoin;
            }
            return false;
        }

        return true;
    }

    public static DirectedAcyclicGraph<Subtask, DefaultEdge> createNFJfromDecompositionTree(
            final BinaryDecompositionTree<Subtask> tree) {
        final DirectedAcyclicGraph<Subtask, DefaultEdge> jobDag =
                new DirectedAcyclicGraph<>(DefaultEdge.class);

        if (tree.getRootNode() != null) {
            traverseNodes(jobDag, tree.getRootNode());
        }

        return jobDag;

    }

    private static GraphEndPoints traverseNodes(final DirectedAcyclicGraph<Subtask, DefaultEdge> jobDag,
            final Node<Subtask> node) {
        switch (node.getNodeType()) {
            case LEAF: {
                final Subtask j = node.getObject();
                jobDag.addVertex(j);
                final Set<Subtask> endPoints = new HashSet<Subtask>(Arrays.asList(j));
                return new GraphEndPoints(endPoints, endPoints);
            }
            case SEQ: {
                final GraphEndPoints leftEndPoints = traverseNodes(jobDag, node.getLeftNode());
                final GraphEndPoints rightEndPoints = traverseNodes(jobDag, node.getRightNode());
                for (final Subtask l : leftEndPoints.right) {
                    for (final Subtask r : rightEndPoints.left) {
                        if (r == l) {
                            continue;
                        }
                        try {
                            jobDag.addDagEdge(l, r);
                        } catch (final Exception e) {
                        }
                    }
                }

                return new GraphEndPoints(leftEndPoints.left, rightEndPoints.right);
            }
            case PAR: {
                final GraphEndPoints leftEndPoints = traverseNodes(jobDag, node.getLeftNode());
                final GraphEndPoints rightEndPoints = traverseNodes(jobDag, node.getRightNode());
                final Set<Subtask> leftEndPointJobs =
                        Sets.newHashSet(Iterables.concat(leftEndPoints.left, rightEndPoints.left));
                final Set<Subtask> rightEndPointJobs = Sets
                        .newHashSet(Iterables.concat(leftEndPoints.right, rightEndPoints.right));
                return new GraphEndPoints(leftEndPointJobs, rightEndPointJobs);
            }
            default:
                break;
        }
        return new GraphEndPoints(new HashSet<>(), new HashSet<>());
    }

    public static class GraphEndPoints {
        public Set<Subtask> left;
        public Set<Subtask> right;

        public GraphEndPoints(final Set<Subtask> left, final Set<Subtask> right) {
            this.left = left;
            this.right = right;
        }
    }
}
